(*Ŀ
  Source                                                                  
                                                                  
                                                                            
 Real Gouraud Shading!                                                      
 Created by Bjarke Viksoe, jan 1996                                         
 E-mail me at: bjarke.viksoe@ntsrv.capacity.dk                              
Ĵ
  Info                                                                    
                                                                    
                                                                            
 Shading really well.                                                       
 Use 3D Studio meshes with many polygons for optimal result.                
                                                                            
 It is surprising how well this CHRMFACE object really looks. Notice that   
 NO modifications has been done to the real object except for removing the  
 eyes and the mouth objects (they were supposed to be rendered black anyway!
                                                                            
                          (Echape pour quitter)                             
Ĵ
  Fichiers ncessaires                                                    
                                                    
                                                                            
 -> 'ChrmFace.3ds' : Face du mec en fichier 3D Studio                       
*)

{$M 4096,0,0}

{-$DEFINE Fps: Affiche le nombre d'image/seconde }

{$S-,R-,N+,G+}

Uses
  Clavier,ModeX,OptVideo,PalVideo,Polic8x8,X3DS {$IFDEF Fps} ,Fps {$ENDIF};

Function LongMul (A,B: Integer): Longint; Inline ($5A/$58/$F7/$EA);
Function LongDiv (A: Longint; B: Integer) : Integer; Inline ($59/$58/$5A/$F7/$F9);

const
  CoulFond = 0;

TYPE
  tNewCoord = RECORD {size is 16 bytes}
    x,y,z : integer;
    nx,ny,nz : integer;
    EdgeNormal : longint;
  end;

  tCoordBuffer = array[0..ARRAY_MAXCOORDS] of tCoordRec;
  tFaceArray = array[0..ARRAY_MAXFACES] of tFaceRec;
  tRotatedCoordBuffer = array[0..ARRAY_MAXCOORDS] of tNewCoord;
  tSlopeType = array[0..320*2] of integer;

var
  Dummy,TableSqrt: array[0..4095] of byte;

  slope, zslope : tSlopeType;
  face          : tFaceArray;          {holds face-definitions}
  coords        : tCoordBuffer;        {holds original coordinates}
  cbuffer       : tRotatedCoordBuffer; {holds rotated coordinates}
  number_faces  : word;
  number_coords : word;

  a,b : longint;
  minx,maxx : integer;

  TableSin: array[0..639] of integer;
  v1,v2,v3   : word;
  cos1,sin1,cos2,sin2,cos3,sin3 : integer;

(*------------------------------------------------*)
(*             INITIALIZE RENDERING               *)
(*------------------------------------------------*)

Procedure SetupSinus;
Var
	i : integer;
	v, vadd : real;
Begin
  v := 0.0;
  vadd := (2.0*pi/512.0);
  for i:=0 to 639 do begin
    TableSin[i] := round(sin(v)*32767);
    v := v +vadd;
  end;
End;


Procedure InitDemo;
Var
	i,j : integer;
	c : longint;
	v,xv, amb, dif, spec : real;
Begin
{	ClearWholeScreen;}
	SetupSinus;

	v1:=0; v2:=0; v3:=128;

	{make a nice blue coloured duck}
	v:=pi / 2;
	xv:=pi / 128;
	for i:=1 to 128 do begin
		amb := 0.0;
		dif := cos(v) * 10.0;
		spec := cos(v) * 120.0;
		c := trunc((amb+dif+spec)/4.0);
		ChangePal( i, 0,c DIV 1,c DIV 2);
		v:=v-xv;
	end;
	ChangePal(0, 0,0,0);
	for i:=128 to 255 do ChangePal(i, 0,0,0);

	j:=0;
	FillChar(Dummy,SizeOf(Dummy),0);
	for i:=0 to 4095 do begin
		if Sqr(j+1)=i then Inc(j);
		TableSqrt[i]:=j;
	end;
End;

(*------------------------------------------------*)
(*            START RENDERING PROCESS             *)
(*------------------------------------------------*)

Procedure FindFaceZ; Assembler;
{Finds the medium z-value for each face-record}
Asm
	lea	si,face+tFaceRec.l1
	lea	di,cbuffer+tNewCoord.z {point a first z-coord in coordbuffer...
											easy to index this way}
	cld
	mov	cx,[number_faces]
@loop:
{ face[i].z := (z1 + z2 + z3) DIV 3; }
	lodsw
	shl	ax,4
	mov	bx,ax
	mov	dx,[di+bx]
	lodsw
	shl	ax,4
	mov	bx,ax
	add	dx,[di+bx]
	sar	dx,1
	lodsw
	shl	ax,4
	mov	bx,ax
	add	dx,[di+bx]
	sar	dx,1
	mov	[si-8],dx {put result in face[i].z}
	add	si,2
	dec	cx
	jnz	@loop
End;

Procedure QuickSort(lo,hi:integer);
 Procedure Sort(l,r:integer);
 Var
	 i,j,x : integer;
 Begin
	i:=l; j:=r;
	x:=face[(l+r) SHR 1].z;
	repeat
	  if i < j then begin
		 while face[i].z < x do inc(i);
		 while x < face[j].z do dec(j);
		 asm
			 lea	si,face	{ swap face records... face[i] <-> face[j] }
			 mov	di,si
			 mov	ax,[i]
			 shl	ax,3
			 add	di,ax
			 mov	ax,[j]
			 shl	ax,3
			 add	si,ax
			 DB EX; mov ax,[di]
			 DB EX; xchg [si],ax
			 DB EX; stosw;
			 DB EX; mov ax,[di]
			 DB EX; xchg [si+4],ax
			 DB EX; stosw;
		 end;
		 inc(i); dec(j);
	  end;
	until i >= j;
	if l < j then Sort(l,j);
	if i < r then Sort(i,r);
 End;
Begin
	asm
		mov	ax,ds {we hope that BP won't change ES for a while...}
		mov	es,ax
		cld
	end;
	Sort(lo,hi);
End;


(*------------------------------------------------*)
(*       CALCULATE FACES POSITION ON SCREEN       *)
(*------------------------------------------------*)

Procedure ClearSlope; assembler; asm
	mov	ax,ds
	mov	es,ax
	lea	di,slope
	DB EX; mov ax,$8000; DW $8000;
	cld
	mov	cx,TYPE(tSlopeType)/4
	rep; DB EX; stosw
End;

Procedure CalcSlope(l1,l2 : word); assembler;
Var
	light1,light2,coladd : word;
	xlowadd : word;
	ysize : integer;
Asm
	lea	si,cbuffer
	DB EX; xor cx,cx
	mov	bx,[l1]							{get first coords}
	shl	bx,4
	mov	ax,WORD PTR [si+bx+tNewCoord.EdgeNormal]	{get light value}
	mov	[light2],ax
	mov	dx,[si+bx+tNewCoord.x]		{get x/y coords}
	mov	cx,[si+bx+tNewCoord.y]

	mov	ax,[l2]							{get second coords}
	shl	ax,4
	add	si,ax
	mov	ax,WORD PTR [si+tNewCoord.EdgeNormal]		{get z value}
	mov	[light1],ax
	mov	ax,[si+tNewCoord.x]			{get x/y coords}
	mov	bx,[si+tNewCoord.y]

	cmp	bx,cx								{make sure we go downwards...}
	jle	@noswap
	mov	si,[light1]						{swap z}
	xchg	[light2],si
	mov	[light1],si
	xchg	ax,dx								{swap x}
	xchg	bx,cx								{sway y}
@noswap:

	or		bx,bx
	js		@zero
	cmp	bx,[minx]						{record miny and maxy}
	jae	@min_x
	mov	[minx],bx
@min_x:
	cmp	cx,[maxx]
	jbe	@max_x
	mov	[maxx],cx
@max_x:

	sub	cx,bx								{find y-size}
	jcxz	@zero
	mov	[ysize],cx
	add	bx,bx
	add	bx,bx
	lea	si,[slope]
	add	si,bx

	push	ax
	sub	dx,ax

	mov	ax,dx								{calc x-slope run}
	DB EX; shl	ax,16
	{cdq} DB EX; cwd
	DB EX; idiv cx
	DB EX; mov dx,ax
	DB EX; shr dx,16
	mov [xlowadd],ax
	{DX also loaded... but kept alive}

	push dx									{also calc z-slope run}
	mov dh,BYTE PTR [light1]
	mov ah,BYTE PTR [light2]
	sub ah,dh
	xor al,al
	cwd
	idiv cx
	mov [coladd],ax
	pop dx
@one:
	pop	cx

	xor	bx,bx
	mov	ah,BYTE PTR [light1] 		{prepare also z-slope calc. light1:=light1*256}
	xor	al,al
	mov	di,$8000
@loop:
	cmp	[si],di							{is first slot filled?}
	jne	@other							{yes, put it in 2nd}
	mov	[si+TYPE(tSlopeType)],ah	{insert z-coord}
	mov	[si],cx							{insert x-coord}
	add	bx,[xlowadd]					{add to x-coord}
	adc	cx,dx
	add	ax,[coladd]						{add to z-coord}
	add	si,4								{prepare next slot...}
	dec	[ysize]
	jnz	@loop
	jmp	NEAR PTR @zero
@other:
	mov	[si+TYPE(tSlopeType)+2],ah {insert in 2nd slot}
	mov	[si+2],cx
	add	bx,[xlowadd]
	adc	cx,dx
	add	ax,[coladd]
	add	si,4
	dec	[ysize]
	jnz	@loop
@zero:
End;


(*------------------------------------------------*)
(*       PREPARE AND DO ROTATION OF POINTS        *)
(*------------------------------------------------*)

Procedure CalcAngle; Begin
  {get sinus values}
  sin1 := TableSin[v1]; cos1 := TableSin[v1+128];
  sin2 := TableSin[v2]; cos2 := TableSin[v2+128];
  sin3 := TableSin[v3]; cos3 := TableSin[v3+128];

  {rotate object a bit}
  v1 := (v1+4) AND 511;
  v2 := (v2-3) AND 511;
  v3 := (v3+2) AND 511;
End;

Procedure RotateAllCoords; assembler;
{Rotate all coords in "coords" around all 3 axis and make
 perspective calcualtion. Store x,y,z results in "cbuffer"}
Var
  xcoord,ycoord,zcoord, n : integer;
Asm
  mov ax,ds
  mov es,ax
  lea si,[coords]
  lea di,[cbuffer]
  mov ax,[number_coords]
  mov [n],ax
  cld
@loop:
  lodsw
  mov [xcoord],ax
  lodsw
  mov [ycoord],ax
  lodsw
  mov [zcoord],ax

  mov ax,[xcoord]             {rotate around Z-axis}
  push ax
  imul [Cos1]
  add ax,ax
  adc dx,dx
  mov bx,dx
  mov ax,[ycoord]
  imul [Sin1]
  add ax,ax
  adc dx,dx
  sub bx,dx
  mov [xcoord],bx
  pop ax
  imul [Sin1]
  add ax,ax
  adc dx,dx
  mov bx,dx
  mov ax,[ycoord]
  imul [Cos1]
  add ax,ax
  adc dx,dx
  add bx,dx
  mov [ycoord],bx

  mov ax,[ycoord]             {rotate around Y-axis}
  push ax
  imul [Cos2]
  add ax,ax
  adc dx,dx
  mov bx,dx
  mov ax,[zcoord]
  imul [Sin2]
  add ax,ax
  adc dx,dx
  sub bx,dx
  mov [ycoord],bx
  pop ax
  imul [Sin2]
  add ax,ax
  adc dx,dx
  mov bx,dx
  mov ax,[zcoord]
  imul [Cos2]
  add ax,ax
  adc dx,dx
  add bx,dx
  mov [zcoord],bx

  mov ax,[xcoord]             {rotate around X-axis}
  push ax
  imul [Cos3]
  add ax,ax
  adc dx,dx
  mov bx,dx
  mov ax,[zcoord]
  imul [Sin3]
  add ax,ax
  adc dx,dx
  sub bx,dx
  mov [xcoord],bx
  pop ax
  imul [Sin3]
  add ax,ax
  adc dx,dx
  mov bx,dx
  mov ax,[zcoord]
  imul [Cos3]
  add ax,ax
  adc dx,dx
  add bx,dx
  mov [zcoord],bx

  add bx,800
  or bx,bx
  jnz @zero
  mov bl,1
@zero:

  mov ax,[xcoord]
  cwd
  mov dl,ah
  mov ah,al
  xor al,al
  idiv bx
  add ax,100
  stosw

  mov ax,[ycoord]
  cwd
  mov dl,ah
  mov ah,al
  xor al,al
  idiv bx
  add ax,160
  stosw

  mov ax,bx
  sub ax,800
  stosw
  DB EX; xor ax,ax
  stosw
  DB EX; stosw
  DB EX; stosw

  dec [n]
  jnz @loop
End;


Function FaceShown(l1,l2,l3 : word) : boolean; Begin asm
  mov cx,[l1] {*16 because a 'cbuffer' record is 16 bytes long}
  shl cx,4
  shl [l2],4
  shl [l3],4
  lea si,cbuffer

{ a := x1-x2 * y3-y2; }
  mov bx,cx
  mov ax,[si+bx]
  mov bx,[l3]
  mov dx,[si+bx+2]
  mov bx,[l2]
  sub ax,[si+bx]
  sub dx,[si+bx+2]
  imul dx
  mov WORD PTR [a+2],ax
  mov WORD PTR [a],dx

{ b := y1-y2 * x3-x2; }
  mov bx,cx
  mov ax,[si+bx+2]
  mov bx,[l3]
  mov dx,[si+bx]
  mov bx,[l2]
  sub ax,[si+bx+2]
  sub dx,[si+bx]
  imul dx

  mov WORD PTR [b+2],ax
  mov WORD PTR [b],dx
  end;
  FaceShown := (a-b) > 0;
End;


Function CalcLight (l1,l2,l3 : word) : integer;
Var
   CX1,CY1,CZ1, CX2,CY2,CZ2 : integer;
   DX,DY,DZ : longint;
   PX,PY,PZ : integer;
   Quadrat : integer;
Begin
  with cbuffer[l1] do begin
    PX := x;
    PY := y;
    PZ := z;
  end;
  CX1 := cbuffer[l2].x-PX;
  CY1 := cbuffer[l2].y-PY;
  CZ1 := cbuffer[l2].z-PZ;
  CX2 := cbuffer[l3].x-PX;
  CY2 := cbuffer[l3].y-PY;
  CZ2 := cbuffer[l3].z-PZ;
  DX := (LongMul(CY1,CZ2)-LongMul(CZ1,CY2));
  DY := (LongMul(CZ1,CX2)-LongMul(CX1,CZ2));
  DZ := (LongMul(CX1,CY2)-LongMul(CY1,CX2));
  with cbuffer[l1] do begin
    Inc( nx,DX );
    Inc( ny,DY );
    Inc( nz,DZ );
  end;
  with cbuffer[l2] do begin
    Inc( nx,DX );
    Inc( ny,DY );
    Inc( nz,DZ );
  end;
  with cbuffer[l3] do begin
    Inc( nx,DX );
    Inc( ny,DY );
    Inc( nz,DZ );
  end;
End;

Procedure FillShape (x,xsize : integer); assembler;
Var
  c1,c2 : byte;
  bitxpos : byte;
Asm
  mov ax,[xsize]
  jz @done
  cmp ax,80
  jae @done
  mov ax,[x]        {find VGA display offset}
  sar ax,2
  add ax,[InfoModX.OfstActu]
  mov di,ax

  lea si,[slope]        {find x1/x2 pos in 'slope' array}
  mov ax,[x]
  mov cx,ax
  shl ax,2
  add si,ax

  and cl,3
  mov al,$11
  shl al,cl
  mov [bitxpos],al

  mov es,[SEGA000]
  mov dx,3C4h
  mov al,02h
  out dx,al
  cld
@xloop:
  mov bh,[si+TYPE(tSlopeType)] {fetch color value}
  lodsw                   {fetch first xpos}
  mov dx,ax
  mov bl,[si+TYPE(tSlopeType)] {fetch second color value}
  lodsw                   {fetch second xpos}
  cmp ax,dx                    {make sure we go downwards...}
  jle @exchange
  xchg ax,dx
  xchg bl,bh
@exchange:
  mov [c1],bl
  mov [c2],bh

  cmp dx,0
  jl @filledout_fast
  cmp ax,200
  jge @filledout_fast
  cmp ax,0
  jge @cut1
  xor ax,ax
@cut1:
  cmp dx,199
  jle @cut2
  mov dx,199
@cut2:
  push si
  push di

  mov bx,ax        {ajust VGA address offset}
  shl bx,4
  add di,bx
  shl bx,2
  add di,bx      { DI = Y*80 }

  mov cx,dx        {find height of line}
  sub cx,ax
  jcxz @filledout

  mov ah,[c2]        {prepare z-slope run}
  sub ah,[c1]
  xor al,al
  cwd
  idiv  cx
  mov bx,ax

  mov dx,$3C5        {set VGA bitplane register}
  mov al,[bitxpos]
  out  dx,al

  mov ah,[c1]          {prepare color-slope run}
  xor  al,dl
  mov dx,80
@loop:
  add ax,bx          {add to color-coord run}
  mov es:[di],ah      {put color on VGA display}
  add di,dx          {find next VGA line}
  dec cl
  jnz @loop

@filledout:
  pop di
  pop si
@filledout_fast:
  rol [bitxpos],1
  adc di,0          {find next x-position}
  dec [xsize]
  jnz @xloop
@done:
End;

Function IntSqrt (const Nbr: LongInt) : Word; Assembler; Asm
  DB 66h; mov ax,WORD PTR [Nbr]
  DB 66h; mov bx,ax
  DB 66h; mov cx,0000h; DW 4000h
  DB 66h; mov dx,cx
  DB 66h; add dx,ax
  DB 66h; shr ax,1
  DB 66h; cmp dx,bx
  ja @over1
  DB 66h; sub bx,dx
  DB 66h; or ax,cx
@over1:
  DB 66h; shr cx,2
  DB 66h; mov dx,cx
  DB 66h; add dx,ax
  DB 66h; shr ax,1
  DB 66h; cmp dx,bx
  ja @over2
  DB 66h; sub bx,dx
  DB 66h; or ax,cx
@over2:
  DB 66h; shr cx,2
  DB 66h; mov dx,cx
  DB 66h; add dx,ax
  DB 66h; shr ax,1
  DB 66h; cmp dx,bx
  ja @over3
  DB 66h; sub bx,dx
  DB 66h; or ax,cx
@over3:
  DB 66h; shr cx,2
  DB 66h; mov dx,cx
  DB 66h; add dx,ax
  DB 66h; shr ax,1
  DB 66h; cmp dx,bx
  ja @over4
  DB 66h; sub bx,dx
  DB 66h; or ax,cx
@over4:
  DB 66h; shr cx,2
  DB 66h; mov dx,cx
  DB 66h; add dx,ax
  DB 66h; shr ax,1
  DB 66h; cmp dx,bx
  ja @over5
  DB 66h; sub bx,dx
  DB 66h; or ax,cx
@over5:
  DB 66h; shr cx,2
  DB 66h; mov dx,cx
  DB 66h; add dx,ax
  DB 66h; shr ax,1
  DB 66h; cmp dx,bx
  ja @over6
  DB 66h; sub bx,dx
  DB 66h; or ax,cx
@over6:
  DB 66h; shr cx,2
  DB 66h; mov dx,cx
  DB 66h; add dx,ax
  DB 66h; shr ax,1
  DB 66h; cmp dx,bx
  ja @over7
  DB 66h; sub bx,dx
  DB 66h; or ax,cx
@over7:
  DB 66h; shr cx,2
  DB 66h; mov dx,cx
  DB 66h; add dx,ax
  DB 66h; shr ax,1
  DB 66h; cmp dx,bx
  ja @over8
  DB 66h; sub bx,dx
  DB 66h; or ax,cx
@over8:
  DB 66h; shr cx,2
  DB 66h; mov dx,cx
  DB 66h; add dx,ax
  DB 66h; shr ax,1
  DB 66h; cmp dx,bx
  ja @over9
  DB 66h; sub bx,dx
  DB 66h; or ax,cx
@over9:
  DB 66h; shr cx,2
  DB 66h; mov dx,cx
  DB 66h; add dx,ax
  DB 66h; shr ax,1
  DB 66h; cmp dx,bx
  ja @over10
  DB 66h; sub bx,dx
  DB 66h; or ax,cx
@over10:
  DB 66h; shr cx,2
  DB 66h; mov dx,cx
  DB 66h; add dx,ax
  DB 66h; shr ax,1
  DB 66h; cmp dx,bx
  ja @over11
  DB 66h; sub bx,dx
  DB 66h; or ax,cx
@over11:
  DB 66h; shr cx,2
  DB 66h; mov dx,cx
  DB 66h; add dx,ax
  DB 66h; shr ax,1
  DB 66h; cmp dx,bx
  ja @over12
  DB 66h; sub bx,dx
  DB 66h; or ax,cx
@over12:
  DB 66h; shr cx,2
  DB 66h; mov dx,cx
  DB 66h; add dx,ax
  DB 66h; shr ax,1
  DB 66h; cmp dx,bx
  ja @over13
  DB 66h; sub bx,dx
  DB 66h; or ax,cx
@over13:
  DB 66h; shr cx,2
  DB 66h; mov dx,cx
  DB 66h; add dx,ax
  DB 66h; shr ax,1
  DB 66h; cmp dx,bx
  ja @over14
  DB 66h; sub bx,dx
  DB 66h; or ax,cx
@over14:
  DB 66h; shr cx,2
  DB 66h; mov dx,cx
  DB 66h; add dx,ax
  DB 66h; shr ax,1
  DB 66h; cmp dx,bx
  ja @over15
  DB 66h; sub bx,dx
  DB 66h; or ax,cx
@over15:
  DB 66h; shr cx,2
  DB 66h; mov dx,cx
  DB 66h; add dx,ax
  DB 66h; shr ax,1
  DB 66h; cmp dx,bx
  ja @over16
  DB 66h; sub bx,dx
  DB 66h; or ax,cx
@over16:
End;

(*------------------------------------------------*)
(*                      MAIN                      *)
(*------------------------------------------------*)

Procedure LanceDemoGouraud;
Var
  i : integer;
  Quotient : longint;
Begin
  repeat
    EffacePageEcranX (CoulFond);

    CalcAngle;
    RotateAllCoords;
    FindFaceZ;
    QuickSort(0,number_faces-1);

    for i := 0 to number_faces-1 do with face[i] do CalcLight(l1,l2,l3);

    for i := 0 to number_coords-1 do with cbuffer[i] do begin
      Quotient := IntSqrt( LongMul(nx,nx)+LongMul(ny,ny)+LongMul(nz,nz) );
      if Quotient=0 then Quotient:=1;
      EdgeNormal := LongDiv((longint(Abs(nz)) SHL 6),Quotient);
    end;

    for i := number_faces-1 downto 0 do begin
      with face[i] do if FaceShown(l1,l2,l3) then begin
         ClearSlope;
         minx := 320; maxx := 0;
         CalcSlope(l1,l2);
         CalcSlope(l2,l3);
         CalcSlope(l3,l1);
         FillShape (minx, maxx-minx);
      end;
    end;

    {$IFDEF Fps} ActuFps; {$ELSE} AttendEcran; {$ENDIF}
    ChangePageVideoX;
  until TouchPresse;
End;

Begin
  {include all objects you wish to render here...}
  {use '*' to include all meshes!}
  IncludeObject[1]:='*';
  IncludeObject[2]:=''; {end with empty string}
  {now extract all coords and faces...}
  {Parameters:
   1. .3ds filename
   2. scale
   rest are internal to renderer...}
  if not Load3dsObject( 'CHRMFACE.3DS', 0.2, face, coords, number_faces, number_coords ) then begin
    Writeln ('Fichier 3D-Studio ''ChrmFace.3DS'' introuvable!');
    halt; {}
  end;

  {do the rendering thing...}
  LanceModeX;

  InitDemo;
  ChangePal(CoulFond,30,30,30);
  {$IFDEF Fps}
    InitFps (0,0,5,TRUE);
  {$ENDIF}

  LanceDemoGouraud;

  ModeTxt;
  {$IFDEF Fps} FinFps; {$ENDIF}
End.
